##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Make example classes available at module scope.
"""
import zope.location.location


class Demo(object):

    _frozen = None

    def isFrozen(self):
        return self._frozen is not None

    def freeze(self):
        self._frozen = Data()


class Data(object):
    pass


class Subobject(zope.location.location.Location):

    def __init__(self):
        self.counter = 0

    def __call__(self):
        res = self.counter
        self.counter += 1
        return res


class Something(object):
    pass


class Other(object):
    root = object()  # immutable
    __name__ = property(lambda _self: 'something')
    __parent__ = property(lambda self: self.__class__.root)



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest


class _Base(object):

    def setUp(self):
        from zope.interface.interface import adapter_hooks
        self._restore = adapter_hooks[:]

    def tearDown(self):
        from zope.interface.interface import adapter_hooks
        adapter_hooks[:] = self._restore


class Test_clone(_Base, unittest.TestCase):

    def _callFUT(self, obj):
        from zope.copy import clone
        return clone(obj)

    def test_wo_hooks(self):
        from zope.copy.examples import Demo
        demo = Demo()
        demo.freeze()
        self.assertTrue(demo.isFrozen())
        copied = self._callFUT(demo)
        self.assertFalse(copied is demo)
        self.assertTrue(isinstance(copied, Demo))
        self.assertTrue(copied.isFrozen())

    def test_w_simple_hook(self):
        from zope.copy.examples import Data
        from zope.copy.examples import Demo
        from zope.copy.interfaces import ICopyHook
        demo = Demo()
        demo.freeze()

        class Hook(object):
            def __init__(self, context):
                self.context = context

            def __call__(self, obj, register):
                return None

        def _adapt(iface, obj):
            if iface is ICopyHook and isinstance(obj, Data):
                return Hook(obj)
        _registerAdapterHook(_adapt)
        copied = self._callFUT(demo)
        self.assertFalse(copied is demo)
        self.assertTrue(isinstance(copied, Demo))
        self.assertFalse(copied.isFrozen())

    def test_subobject_wo_post_copy_hook(self):
        from zope.location.location import Location
        from zope.location.location import locate

        from zope.copy.examples import Subobject
        o = Location()
        s = Subobject()
        o.subobject = s
        locate(s, o, 'subobject')
        self.assertTrue(s.__parent__ is o)
        self.assertEqual(o.subobject(), 0)
        self.assertEqual(o.subobject(), 1)
        self.assertEqual(o.subobject(), 2)
        c = self._callFUT(o)
        self.assertTrue(c.subobject.__parent__ is c)
        self.assertEqual(c.subobject(), 3)
        self.assertEqual(o.subobject(), 3)

    def test_subobject_w_post_copy_hook(self):
        from zope.location.location import Location
        from zope.location.location import locate

        from zope.copy.examples import Subobject
        from zope.copy.interfaces import ICopyHook
        o = Location()
        s = Subobject()
        o.subobject = s
        locate(s, o, 'subobject')
        self.assertTrue(s.__parent__ is o)
        self.assertEqual(o.subobject(), 0)
        self.assertEqual(o.subobject(), 1)
        self.assertEqual(o.subobject(), 2)

        class Hook(object):
            def __init__(self, context):
                self.context = context

            def __call__(self, obj, register):
                obj = Subobject()

                def reparent(translate):
                    obj.__parent__ = translate(self.context.__parent__)
                register(reparent)
                return obj

        def _adapt(iface, obj):
            if iface is ICopyHook and isinstance(obj, Subobject):
                return Hook(obj)
        _registerAdapterHook(_adapt)
        c = self._callFUT(o)
        self.assertTrue(c.subobject.__parent__ is c)
        self.assertEqual(c.subobject(), 0)
        self.assertEqual(o.subobject(), 3)


class Test_copy(_Base, unittest.TestCase):

    def _callFUT(self, obj):
        from zope.copy import copy
        return copy(obj)

    def test_clears_attrs(self):
        from zope.copy.examples import Demo
        parent = Demo()
        demo = Demo()
        demo.__parent__ = parent
        demo.__name__ = 'demo'
        copied = self._callFUT(demo)
        self.assertFalse(copied is demo)
        self.assertTrue(isinstance(copied, Demo))
        self.assertEqual(copied.__parent__, None)
        self.assertEqual(copied.__name__, None)

    def test_w_readonly___parent___and___name__(self):
        global Foo  # make unpicklable
        parent = object()

        class Foo(object):
            @property
            def __parent__(self):
                return parent

            @property
            def __name__(self):
                return 'foo'
        foo = Foo()
        copied = self._callFUT(foo)
        self.assertFalse(copied is foo)
        self.assertTrue(isinstance(copied, Foo))
        self.assertTrue(copied.__parent__ is parent)
        self.assertEqual(copied.__name__, 'foo')


class CopyPersistentTests(_Base, unittest.TestCase):

    def _getTargetClass(self):
        from zope.copy import CopyPersistent
        return CopyPersistent

    def _makeOne(self, obj):
        return self._getTargetClass()(obj)

    def test_ctor(self):
        obj = object()
        cp = self._makeOne(obj)
        self.assertTrue(cp.toplevel is obj)
        self.assertEqual(cp.pids_by_id, {})
        self.assertEqual(cp.others_by_pid, {})
        self.assertEqual(cp.registered, [])

    def test_id_wo_hook(self):
        obj = object()
        cp = self._makeOne(obj)
        self.assertEqual(cp.id(obj), None)

    def test_id_w_hook_already_cached(self):
        from zope.copy.interfaces import ICopyHook
        obj = object()
        cp = self._makeOne(obj)
        cp.pids_by_id[id(obj)] = 'PID'

        class Hook(object):
            def __init__(self, context):
                self.context = context

            def __call__(self, obj, register):
                raise AssertionError("Not called")

        def _adapt(iface, obj):
            assert iface is ICopyHook
            return Hook(obj)
        _registerAdapterHook(_adapt)
        self.assertEqual(cp.id(obj), 'PID')

    def test_id_w_hook_raising_ResumeCopy(self):
        from zope.copy.interfaces import ICopyHook
        from zope.copy.interfaces import ResumeCopy
        obj = object()
        cp = self._makeOne(obj)

        class Hook(object):
            def __init__(self, context):
                self.context = context

            def __call__(self, obj, register):
                raise ResumeCopy()

        def _adapt(iface, obj):
            assert iface is ICopyHook
            return Hook(obj)
        _registerAdapterHook(_adapt)
        self.assertEqual(cp.id(obj), None)

    def test_id_w_hook_normal(self):
        from zope.copy.interfaces import ICopyHook
        obj = object()
        cp = self._makeOne(obj)

        class Hook(object):
            def __init__(self, context):
                self.context = context

            def __call__(self, obj, register):
                return None

        def _adapt(iface, obj):
            assert iface is ICopyHook
            return Hook(obj)
        _registerAdapterHook(_adapt)
        self.assertEqual(cp.id(obj), 1)
        obj2 = object()
        self.assertEqual(cp.id(obj2), 2)
        self.assertEqual(cp.pids_by_id, {id(obj): 1, id(obj2): 2})
        self.assertEqual(cp.others_by_pid, {1: None, 2: None})


def _registerAdapterHook(func):
    from zope.interface.interface import adapter_hooks
    adapter_hooks.insert(0, func)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import sys


PY3 = sys.version_info[0] >= 3

try:
    from cPickle import Pickler
    from cPickle import Unpickler  # pragma: no cover
except ImportError:
    from pickle import Pickler  # noqa: F401 imported but unused
    from pickle import Unpickler  # noqa: F401 imported but unused

if PY3:
    def _memo(pickler):
        # Python 3 uses a "PicklerMemoProxy" which is not subscriptable
        # by itself
        return pickler.memo.copy()
else:  # pragma: no cover
    from operator import attrgetter

    _memo = attrgetter('memo')


def _get_pid(pickler, oid):
    return _memo(pickler)[oid][0]


def _get_obj(unpickler, pid):
    return _memo(unpickler)[pid]



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""API for zope.copy
"""
import zope.interface


class ResumeCopy(Exception):
    """Don't use the hook, resume the copy.

    This is a special exception, raised from the copy hook to signal
    copier that it should continue copying the object recursively.

    See ICopyHook.__call__ method documentation.
    """


class ICopyHook(zope.interface.Interface):
    """An adapter to an object that is being copied"""

    def __call__(toplevel, register):
        """Given the top-level object that is being copied, return the
        version of the adapted object that should be used in the new copy.

        Raising ResumeCopy means that you are foregoing the hook: the
        adapted object will continue to be recursively copied as usual.

        If you need to have a post-copy actions executed, register a
        callable with ``register``.  This callable must take a single
        argument: a callable that, given an object from the original,
        returns the equivalent in the copy.

        See README.txt for more explanation.
        """



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import tempfile

from zope.copy import interfaces
from zope.copy._compat import Pickler
from zope.copy._compat import Unpickler
from zope.copy._compat import _get_obj
from zope.copy._compat import _get_pid


def clone(obj):
    """Clone an object by pickling and unpickling it"""
    with tempfile.TemporaryFile() as tmp:
        persistent = CopyPersistent(obj)

        # Pickle the object to a temporary file
        pickler = Pickler(tmp, protocol=-1)
        pickler.persistent_id = persistent.id
        pickler.dump(obj)

        # Now load it back
        tmp.seek(0)
        unpickler = Unpickler(tmp)
        unpickler.persistent_load = persistent.load

        res = unpickler.load()

        # run the registered cleanups
        def convert(obj):
            pid = _get_pid(pickler, id(obj))
            try:
                return _get_obj(unpickler, pid)
            except KeyError:  # pragma: no cover (PyPy)
                return _get_obj(unpickler, str(pid))
        for call in persistent.registered:
            call(convert)
        return res


def copy(obj):
    """Clone an object, clearing the __name__ and __parent__ attribute
    values of the copy."""
    res = clone(obj)
    if getattr(res, '__parent__', None) is not None:
        try:
            res.__parent__ = None
        except AttributeError:
            pass
    if getattr(res, '__name__', None) is not None:
        try:
            res.__name__ = None
        except AttributeError:
            pass
    return res


class CopyPersistent(object):
    """A helper class providing the persisntent_id and persistent_load
    functions for pickling and unpickling respectively.

    It uses the adaptation to ICopyHook to allow control over object
    copying. See README.txt for more information on that mechanism.
    """

    def __init__(self, obj):
        self.toplevel = obj
        self.pids_by_id = {}
        self.others_by_pid = {}
        self.load = self.others_by_pid.get
        self.registered = []

    def id(self, obj):
        hook = interfaces.ICopyHook(obj, None)
        if hook is not None:
            oid = id(obj)
            if oid in self.pids_by_id:
                return self.pids_by_id[oid]
            try:
                res = hook(self.toplevel, self.registered.append)
            except interfaces.ResumeCopy:
                pass
            else:
                pid = len(self.others_by_pid)

                # The following is needed to overcome a bug
                # in pickle.py. The pickle checks the boolean value
                # of the id, rather than whether it is None.
                pid += 1

                self.pids_by_id[oid] = pid
                self.others_by_pid[pid] = res
                return pid
        return None
